
class SidebarPartField:
    """
    Performs introspection into document part fields.
    """

    def __init__(
        self,
        document,
        model,
        field_name,
        options={}
    ):
        """
        `document` = papermerge.core.models.document instance
        `model` = document part model class as provided by 3rd party app
        `field_name` name of the field we are interested in
        """

        self.document = document
        self.model = model
        self.field_name = field_name
        self.options = options

        fields = [
            field
            for field in self.model._meta.fields
            if field.name == field_name
        ]
        if fields:
            # django field matching given field_name
            self.field = fields[0]

    def to_json(self):
        ret = {}

        internal_type = self.get_internal_type()
        ret['class'] = internal_type

        if internal_type == 'ForeignKey':
            r_obj = self.get_value()

            if self.options and self.options[self.field_name]:
                opts = self.options[self.field_name]
                # choice_fields: ['id', 'name'] option instructs
                # that this foreign key must be displayed as
                # choices. Thus, returned keys are:
                # * value
                # * choices
                if 'choice_fields' in opts:
                    value, choices = self._choice_fields_opts(
                        opts=opts,
                        r_obj=r_obj
                    )
                    ret['value'] = value
                    ret['choices'] = choices
                    ret['field_name'] = self.field_name
                elif 'fields' in opts:
                    value = self._fields_opts(
                        opts=opts,
                        r_obj=r_obj
                    )
                    ret['field_name'] = self.field_name
                    ret['value'] = value

            else:
                _f = self.field_name
                msg = f"Field {_f} is foreignkey. You provide field_options"
                raise ValueError(
                    msg
                )
        else:
            ret['value'] = self.get_value()
            ret['field_name'] = self.field_name

        return ret

    def get_internal_type(self):
        return self.field.get_internal_type()

    def get_value(self):
        parts = getattr(self.document, 'parts')
        value = getattr(parts, self.field_name)
        return value

    def _fields_opts(
        self,
        opts,
        r_obj
    ):
        ret_dict = {}
        fields = opts['fields']

        for field in fields:
            ret_dict[field] = getattr(r_obj, field)

        return ret_dict

    def _choice_fields_opts(
        self,
        opts,
        r_obj
    ):
        ret_value = None
        ret_choices = []
        choice_fields = opts['choice_fields']

        if choice_fields and r_obj:
            ret_value = (
                getattr(r_obj, choice_fields[0]),
                getattr(r_obj, choice_fields[1])
            )
        remote_model_objects = getattr(
            self.field.remote_field.model, 'objects'
        )
        for r_model_inst in remote_model_objects.all():
            if choice_fields and r_model_inst:
                ret_choices.append(
                    (
                        getattr(r_model_inst, choice_fields[0]),
                        getattr(r_model_inst, choice_fields[1]),
                    )
                )

        return ret_value, ret_choices


class SidebarPart:
    """
    Wrapper class for managing/rendering document parts
    on sidebar.
    """

    fields = None
    exclude = None
    readonly_fields = ()
    # model class of the document part
    model = None
    field_options = {}

    def __init__(self, document):
        # papermerge.core.models.document instance
        self.document = document

        if not self.model:
            raise ValueError("SidebarPart: missing model attribute")
        self.opts = self.model._meta

    def to_json(self):
        fields = [
            SidebarPartField(
                document=self.document,
                field_name=field,
                model=self.model,
                options=self.field_options
            ).to_json()
            for field in self.fields
        ]
        ret = {
            'label': self.get_label(),
            'verbose_name': self.get_verbose_name(),
            'js_widget': self.get_js_widget(),
            'fields': fields
        }
        return ret

    def get_label(self):
        if hasattr(self, 'label'):
            return getattr(self, 'label')

        return self.opts.app_config.label

    def get_verbose_name(self):
        if hasattr(self, 'verbose_name'):
            return getattr(self, 'verbose_name')

        return self.opts.app_config.verbose_name

    def get_js_widget(self):

        if hasattr(self, 'js_widget'):
            return getattr(self, 'js_widget')

        return 'DefaultWidget'




import os
import subprocess
import logging


logger = logging.getLogger(__name__)


class PreviewCoord:
    """
    Groups page number and height (expressed in pixels).

    The idea is that every document preview, asks for specific page
    number and end pixel dimentions of preview. Because these two values
    are so much repeated - they are better regarded as one coordonate system.

    page_count - is optional argument. Is might be used for formatting
            purposes e.g. if there are 32 pages, then 9th page will be
            formatted with as 'page-09' instead of page-9;
            this is peculiarity is result of pdftoppm utility functionality.
    """

    def __init__(
            self,
            page,
            height,
            step,
            min_height,
            max_height,
            page_count=None
    ):

        for value in (page, height):
            self.basic_positive_validation(value)

        self.page = page
        self.min_height = min_height
        self.max_height = max_height
        self.step = step
        self.height = self.normalize_h(height)
        self.page_count = page_count

    def __str__(self):
        return "({}/{}, {})".format(
            self.page,
            self.page_count,
            self.height
        )

    def basic_positive_validation(self, value):
        err_msg = "{} arg must be a non negative int".format(value)

        if not isinstance(value, int):
            raise ValueError(err_msg)

        if value < 0:
            raise ValueError(err_msg)

    def normalize_h(self, height):
        """
        Given a random height value - return the closest valid
        height i.e. from (min_height, min_height + step, ..., max_height)
        """

        step = self.step
        min_height = self.min_height
        max_height = self.max_height

        if height < min_height or height > max_height:
            return InvalidHeight("Out of bound image height value")

        for item_height in range(min_height, max_height + step, step):
            if abs(item_height - height) <= step / 2:
                return item_height


class InvalidHeight(Exception):
    pass


class Preview:
    """
    Generates on the fly previews of a given document.

    # Task is any callable which take as parameters a tuple of
    # strings - first element of which is utility and next ones
    # are utility command line options.
    """

    PREVIEW_EXT = 'jpg'

    def __init__(
        self,
        document_file,
        # It is convenient (in tests) to have a preview
        # object without a task.
        task=None
    ):
        self.document_file = document_file
        self.task = task

    def ppmroot(self, coord):
        """
        Returns PPM-root of document file of given coord.

        PPM-root is defined as term used by pdftoppm utility
        from poppler package. Anyway, it is "template" path for generating
        output files. Let's say we want to extract jpeg page 1 and 2 of
        the document with name XYZ.pdf. We need to answer first two questions:

        * (1) Where to extract ?
        * (2) What will be the name of newly created files?

        PPM-root answers both answers above. A ppm-root of form /A/B/C/nana
        will extract (1) extract files to /A/B/C/ directory and (2) will
        name them as follows:

            /A/B/C/nana-1.jpeg
            /A/B/C/nana-2.jpeg

        PPM-root is part of absolute filename up to "-<number>.<ext>" .
        """

        root = self.document_file.rootname

        return os.path.join(
            self.document_file.dir_path,
            str(coord.height),
            "{}-page".format(root)
        )

    def abspath(self, coord):
        """
        Absolute path of given coord. The format chosen here is
        very dependent on the pdftoppm (part of poppler package)
        utility.
        """
        ppmroot = self.ppmroot(coord)
        logger.debug(
            "ppmroot={root} coord.page_count={count} coord.page={page}".format(
                root=ppmroot,
                count=coord.page_count,
                page=coord.page
            )
        )

        if coord.page_count <= 9:
            fmt_page = "{root}-{num:d}.{ext}"
        elif coord.page_count > 9 and coord.page_count < 100:
            fmt_page = "{root}-{num:02d}.{ext}"
        elif coord.page_count > 100:
            fmt_page = "{root}-{num:003d}.{ext}"

        returned_value = fmt_page.format(
            root=ppmroot, num=int(coord.page), ext=self.PREVIEW_EXT
        )

        logger.debug(
            "abspath={}".format(returned_value)
        )

        return returned_value

    def is_available(self, coord):

        str_abs_path = self.abspath(coord)

        if not os.path.isfile(str_abs_path):
            return False

        try:
            with open(str_abs_path, "rb") as f:
                return True
        except IOError:
            return False

    def breed_preview(self, coord, validate_paths=True):

        if not self.is_available(coord):
            if not self.task:
                msg = "Preview object doesn't have a task object."
                raise ValueError(msg)

            try:
                if self.document_file.is_image:
                    self.task(
                        *self.image_preview_cmdline(
                            coord, validate_paths=validate_paths
                        )
                    )
                else:
                    self.task(
                        *self.offspring_cmdline(
                            coord, validate_paths=validate_paths
                        )
                    )
            except subprocess.CalledProcessError as e:
                print(e.stderr)
                raise

            if not self.is_available(coord):
                logger.error("Preview generating failed")
                # TODO: Enhacement: in case preview fails, return
                # some generic 200px wide thumbnail

        return self.abspath(coord)

    def image_preview_cmdline(
            self,
            coord,
            validate_paths=False
    ):
        """
        Returns a list of string values in same format as accepted
        by convert utility(the one bundled in imagemagick).
        First item in the list if always name
        of utlity iteself (convert).

        When recreate_paths is True - paths given as arguments to convert will
        be checked if they exists - if they don't - a warning will be issued
        and paths will be created.
        """
        input_file_path = self.document_file.abspath
        output_dir_path = os.path.dirname(self.abspath(coord))

        if validate_paths:
            if not self.document_file.exists:
                raise ValueError(
                    "Breeding preview of non-existing doc '{}' failed.".format(
                        input_file_path
                    )
                )
            if not os.path.exists(output_dir_path):
                os.makedirs(output_dir_path)

        ret = (
            "/usr/bin/convert",
            # input file path
            input_file_path,
            # in pdftoppm utility,
            # size is specified with --scale-to Height
            # In convert utility size is specified as WxH :)
            # Thus - resize xH - scale to have height = H
            "-resize",
            "x{}".format(coord.height),
            # output directory path
            "{root}-{num:d}.{ext}".format(
                root=self.ppmroot(coord),
                num=1,
                ext=self.PREVIEW_EXT
            ),
        )

        return ret

    def offspring_cmdline(
        self,
        coord,
        validate_paths=False
    ):
        """
        Returns a list of string values in same format as accepted
        by pdftoppm utility/offspring. First item in the list if always name
        of utlity iteself (pdftoppm).

        When recreate_paths is True - paths given as arguments to pdftoppm will
        be checked if they exists - if they don't - a warning will be issued
        and paths will be created.
        """
        input_file_path = self.document_file.abspath
        output_dir_path = os.path.dirname(self.abspath(coord))

        if validate_paths:
            if not os.path.exists(input_file_path):
                raise ValueError(
                    "Breeding preview of non-existing doc '{}' failed.".format(
                        input_file_path
                    )
                )
            if not os.path.exists(output_dir_path):
                os.makedirs(output_dir_path, exist_ok=True)

        # for thumbnail images, poor jpeg qulity
        # is acceptable. We save same space this way.
        if coord.height < 300:
            quality = 40
        else:
            quality = 90
        # each argument must be a separate entry in array
        # https://stackoverflow.com/questions/53628699/subprocess-run-simple-scenario-fails
        ret = (
            "pdftoppm",
            "-jpeg",
            "-f",
            str(coord.page),
            "-l",  # generate only one page
            str(coord.page),
            "-scale-to",
            str(coord.height),
            "-jpegopt",
            "quality={}".format(quality),
            # input file path
            input_file_path,
            # output directory path
            self.ppmroot(coord),
        )

        return ret



# Generated by Django 3.0.6 on 2020-06-01 18:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0007_auto_20200419_0746'),
    ]

    operations = [
        #migrations.RemoveField(
        #    model_name='page',
        #    name='text_deu',
        #),
        #migrations.RemoveField(
        #    model_name='page',
        #    name='text_eng',
        #),
        #migrations.RemoveField(
        #    model_name='page',
        #    name='text_fts',
        #),
        migrations.AddField(
            model_name='page',
            name='norm_breadcrump',
            field=models.CharField(default='', max_length=1024),
        ),
        migrations.AddField(
            model_name='page',
            name='norm_doc_title',
            field=models.CharField(default='', max_length=200),
        ),
        migrations.AddField(
            model_name='page',
            name='norm_folder_title',
            field=models.CharField(default='', max_length=200),
        ),
        migrations.AddField(
            model_name='page',
            name='norm_text',
            field=models.TextField(default=''),
        ),
    ]



# Generated by Django 3.0.2 on 2020-04-19 06:39

# import django.contrib.postgres.search
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_remove_document_celery_task_id'),
    ]

    operations = [
#        migrations.AddField(
#            model_name='basetreenode',
#            name='ancestors_fts',
#            field=django.contrib.postgres.search.SearchVectorField(null=True),
#        ),
#        migrations.AddField(
#            model_name='basetreenode',
#            name='title_fts',
#            field=django.contrib.postgres.search.SearchVectorField(null=True),
#        ),
#        migrations.AddField(
#            model_name='document',
#            name='text_fts',
#            field=django.contrib.postgres.search.SearchVectorField(null=True),
#        ),
#        migrations.AddField(
#            model_name='page',
#            name='text_fts',
#            field=django.contrib.postgres.search.SearchVectorField(null=True),
#        ),
    ]



from django.views import defaults


def bad_request_400_custom(
    request,
    exception,
    template_name='core/400.html'
):
    return defaults.bad_request(
        request=request,
        exception=exception,
        template_name=template_name
    )


def permission_denied_403_custom(
        request,
        exception,
        template_name='core/403.html'
):
    return defaults.permission_denied(
        request=request,
        exception=exception,
        template_name=template_name
    )


def page_not_found_404_custom(
    request,
    exception,
    template_name='core/404.html'
):
    return defaults.page_not_found(
        request=request,
        exception=exception,
        template_name=template_name
    )


def server_error_500_custom(
    request,
    template_name='core/500.html'
):
    return defaults.server_error(
        request=request,
        template_name=template_name
    )



# Generated by Django 3.0.8 on 2020-08-28 08:47

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0018_auto_20200721_1320'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='page',
            options={'ordering': ['number']},
        ),
        migrations.AlterField(
            model_name='automate',
            name='dst_folder',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='core.Folder', verbose_name='Destination Folder'),
        ),
        migrations.AlterField(
            model_name='automate',
            name='extract_page',
            field=models.BooleanField(default=False, verbose_name='Extract Page'),
        ),
        migrations.AlterField(
            model_name='automate',
            name='is_case_sensitive',
            field=models.BooleanField(default=True, verbose_name='Is case sensitive'),
        ),
        migrations.AlterField(
            model_name='automate',
            name='match',
            field=models.CharField(blank=True, max_length=256, verbose_name='Match'),
        ),
        migrations.AlterField(
            model_name='automate',
            name='matching_algorithm',
            field=models.PositiveIntegerField(choices=[(1, 'Any'), (2, 'All'), (3, 'Literal'), (4, 'Regular Expression')], default=1, verbose_name='Matching Algorithm'),
        ),
        migrations.AlterField(
            model_name='automate',
            name='name',
            field=models.CharField(max_length=128, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='automate',
            name='plugin_name',
            field=models.CharField(blank=True, choices=[], default=None, max_length=256, null=True, verbose_name='Plugin Name'),
        ),
    ]



from django.utils.translation import ugettext_lazy as _
from django.views import generic
from django.urls import reverse_lazy
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect

from papermerge.contrib.admin.views import mixins as mix
from papermerge.contrib.admin.forms import AutomateForm
from papermerge.core.models import Automate
from papermerge.core.models.folder import get_inbox_children
from papermerge.core.models.page import get_pages


class AutomatesView(LoginRequiredMixin):
    model = Automate
    form_class = AutomateForm
    success_url = reverse_lazy('admin:automates')


class AutomatesListView(
    AutomatesView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):

    title = _("Automates")

    def get_queryset(self):
        qs = super().get_queryset()

        return qs.filter(
            user=self.request.user
        ).order_by('name')

    def post(self, request):

        selected_action = request.POST.getlist('_selected_action')

        go_action = request.POST['action']

        if go_action == 'delete_selected':
            self.delete_selected_entries(selected_action)

        if go_action == 'run_selected':
            self.run_selected_automates(selected_action)

        return redirect(self.success_url)

    def run_selected_automates(self, selection):
        inbox_nodes = get_inbox_children(
            self.request.user
        )

        # Run automates over all page objects from user's inbox.
        # Only pages with non empty text field are taken into account
        self.get_queryset().filter(
            id__in=selection
        ).run(
            get_pages(
                inbox_nodes,
                include_pages_with_empty_text=False
            )
        )


class AutomateCreateView(AutomatesView, generic.CreateView):

    def get_form_kwargs(self):
        """Return the keyword arguments for instantiating the form."""
        kwargs = super().get_form_kwargs()
        kwargs['user'] = self.request.user

        return kwargs

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('New')
        context['action_url'] = reverse_lazy('admin:automate-add')

        return context

    def form_valid(self, form):
        form.instance.user = self.request.user

        return super().form_valid(form)


class AutomateUpdateView(AutomatesView, generic.UpdateView):

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('Edit')
        context['action_url'] = reverse_lazy(
            'admin:automate-update',
            args=(self.object.pk,)
        )

        return context



# Generated by Django 3.0.6 on 2020-06-02 12:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0009_kvstore_kvstorenode_kvstorepage'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='kvstore',
            name='human_key',
        ),
    ]



from django.views.generic.base import ContextMixin
from django.core.paginator import Paginator
from django.shortcuts import redirect
from django.utils.translation import gettext as _
from django.contrib import messages


class RequiredPermissionMixin:

    def dispatch(self, request, *args, **kwargs):

        if not hasattr(self, 'required_permission'):
            return self.handle_no_permission()

        req_permission = getattr(self, 'required_permission')

        if not self.request.user.has_perm(req_permission):
            return self.handle_no_permission()

        return super().dispatch(request, *args, **kwargs)


class CommonListMixin(ContextMixin):
    """
    Provides common context for admin/object_list.html
    """

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)

        context['action_url'] = self.action_url
        context['success_url'] = self.success_url
        context['new_object_url'] = self.new_object_url
        context['title'] = self.title
        context['delete_selected_msg'] = self.get_delete_selected_msg()
        context['empty_list_msg'] = self.get_empty_list_msg()
        context['model_name_title'] = self.get_model_name_plu().capitalize()

        return context

    def get_model_name_plu(self):
        return self.model._meta.verbose_name_plural

    def get_delete_selected_msg(self):

        msg = _("Delete selected %(model_name_plu)s") % {
            'model_name_plu': self.get_model_name_plu()
        }

        return msg

    def get_empty_list_msg(self):

        msg = _("No %(model_name_plu)s so far") % {
            'model_name_plu': self.model._meta.verbose_name_plural
        }

        return msg


class DeleteEntriesMixin:
    """
    Handles HTTP POST with 'delete selected'.
    """

    def get_delete_entries(self, selection):
        return self.get_queryset().filter(
            id__in=selection
        )

    def delete_selected_entries(self, selected_action):
        deleted, row_count = self.get_delete_entries(
            selection=selected_action
        ).delete()

        model_label = self.model._meta.label
        count = row_count.get(model_label, 0)
        msg = _(
            "%(count)s selected entries were deleted."
        ) % {
            'count': count
        }
        messages.info(self.request, msg)

    def post(self, request):
        """
        Delete selected entries
        """
        selected_action = request.POST.getlist('_selected_action')

        go_action = request.POST['action']
        if go_action == 'delete_selected':
            self.delete_selected_entries(selected_action)

        return redirect(self.success_url)


class PaginationMixin:

    PER_PAGE = 10

    def get_context_data(self, **kwargs):

        objs = self.get_queryset()
        page_number = int(self.request.GET.get('page', 1))

        paginator = Paginator(objs, per_page=self.PER_PAGE)
        num_pages = paginator.num_pages
        page_obj = paginator.get_page(page_number)

        # 1.   Number of pages < 7: show all pages;
        # 2.   Current page <= 4: show first 7 pages;
        # 3.   Current page > 4 and < (number of pages - 4): show current page,
        #       3 before and 3 after;
        # 4.   Current page >= (number of pages - 4): show the last 7 pages.

        if num_pages <= 7 or page_number <= 4:  # case 1 and 2
            pages = [x for x in range(1, min(num_pages + 1, 7))]
        elif page_number > num_pages - 4:  # case 4
            pages = [x for x in range(num_pages - 6, num_pages + 1)]
        else:  # case 3
            pages = [x for x in range(page_number - 3, page_number + 4)]

        context = super().get_context_data(**kwargs)
        context.update({
            'object_list': page_obj.object_list,
            'pages': pages,
            'num_pages': num_pages,
            'page_number': page_number,
            'page': page_obj,
            # by the way, add title to the context as well
            'title': self.title
        })

        return context



from django.utils.html import escape


def sanitize_kvstore(kvstore_dict):
    """
    Creates a sanitized dictionary.

    Sanitizied dictionary contains only allowed keys and escaped values.
    """
    allowed_keys = [
        'id',
        'key',
        'value',
        'kv_type',
        'kv_format',
        'kv_inherited',
    ]

    sanitized_kvstore_dict = {}

    for allowed_key in allowed_keys:
        if allowed_key in kvstore_dict.keys():
            value = kvstore_dict.get(allowed_key, None)
            if isinstance(value, bool):
                allowed_value = value
            else:
                allowed_value = escape(kvstore_dict.get(allowed_key, None))

            sanitized_kvstore_dict[allowed_key] = allowed_value

    return sanitized_kvstore_dict


def sanitize_kvstore_list(kvstore_list):
    """
    Creates a new list of sanitized dictionaries.

    Sanitizied dictionary contains only allowed keys and escaped values.
    """
    if not isinstance(kvstore_list, list):
        raise ValueError("Expects list type as input")

    new_kvstore_list = [
        sanitize_kvstore(item) for item in kvstore_list
    ]

    return new_kvstore_list



from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import ugettext_lazy as _
from django.views import generic
from django.urls import reverse_lazy
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.exceptions import PermissionDenied

from papermerge.core.models import User
from papermerge.contrib.admin.forms import (
    UserFormWithoutPassword,
    UserChangePasswordForm
)
from papermerge.contrib.admin.views import mixins as mix


class UsersView(
    LoginRequiredMixin,
    mix.RequiredPermissionMixin
):
    # only superuser can access this view
    only_superuser = True
    model = User
    form_class = UserFormWithoutPassword
    success_url = reverse_lazy('admin:users')

    def dispatch(self, request, *args, **kwargs):

        if self.request.user.is_authenticated:
            if not self._is_allowed(request):
                raise PermissionDenied()

        ret = super().dispatch(request, *args, **kwargs)

        return ret

    def _is_allowed(self, request):
        if hasattr(self, 'only_superuser'):
            if getattr(self, 'only_superuser'):
                return request.user.is_superuser

        return True


class UsersListView(
    UsersView,
    mix.PaginationMixin,
    mix.DeleteEntriesMixin,
    generic.ListView
):
    title = _("Users")
    required_permission = 'core.view_user'

    def get_queryset(self):
        qs = super().get_queryset()

        return qs.order_by('username')


class UserCreateView(UsersView, generic.CreateView):

    required_permission = 'core.add_user'

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('New')
        context['action_url'] = reverse_lazy('admin:user-add')

        return context


class UserUpdateView(UsersView, generic.UpdateView):

    required_permission = 'core.change_user'

    def get_context_data(self, **kwargs):

        context = super().get_context_data(**kwargs)
        context['title'] = _('Edit')
        context['action_url'] = reverse_lazy(
            'admin:user-update',
            args=(self.object.pk,)
        )

        return context


class UserChangePasswordView(UsersView, generic.FormView):

    required_permission = 'core.change_user'
    form_class = UserChangePasswordForm
    template_name = 'core/user_form.html'

    def get_context_data(self, **kwargs):

        user = User.objects.get(id=self.kwargs['pk'])
        context = super().get_context_data(**kwargs)
        context['title'] = _('Change Password')
        context['object'] = user
        context['action_url'] = reverse_lazy(
            'admin:user-change-password',
            args=(user.pk,)
        )

        return context

    def form_valid(self, form):
        password = form.cleaned_data['password1']
        user = User.objects.get(id=self.kwargs['pk'])
        user.set_password(password)
        user.save()
        messages.success(
            self.request,
            _(
                "%(username)s password was successfully updated!"
            ) % {
                'username': user.username
            }
        )
        return redirect(self.success_url)



import json
from functools import wraps
from django.http import (
    HttpResponse,
    HttpResponseRedirect,
    HttpResponseForbidden
)
from django.utils.log import log_response


def smart_dump(value):

    if isinstance(value, str):
        return json.dumps({'msg': value})

    if isinstance(value, dict):
        return json.dumps(value)

    return ""


def json_response(func):
    """
    Decorates view to return application/json type response.
    Argument function func is expected to return one of:

        1. A string
            in this case, body will be a json.dump({
                'msg': returned_str
            })
            and status code will be 200
        2. A dictionary
            same as above, but respone will dump directly dictionary
                json.dumps(returned_dict)
            and status code will be 200
        3. Two valued tuple
            First value of the tuple must be either a string
            or a dictionary. In this case above points 1 and 2 apply
            Second value of the tuple is status code (as intiger number)
    """
    def inner(*args, **kwargs):
        ret = func(*args, **kwargs)
        status = 200
        body = ""

        if isinstance(ret, str) or isinstance(ret, dict):
            body = smart_dump(ret)
        elif isinstance(ret, tuple):
            for_body = ret[0]
            status = ret[1]
            body = smart_dump(for_body)
        elif isinstance(ret, HttpResponseRedirect):
            # in case anonymous user access this view - return
            # the HttpResponseRedirect object
            return ret
        else:
            raise ValueError(
                "Function must return str, dict or 2 valued tuple"
            )

        return HttpResponse(
            body,
            content_type="application/json",
            status=status
        )

    return inner


def require_PERM(perm):
    """
    Decorator to make a view only accept users which have given permission.
    Usage:

        @require_PERM('core.add_folder')
        def my_view(request):
            # Can assume now that logged in user has 'core.add_folder'
            # permission
            # ...
    """
    def decorator(func):
        @wraps(func)
        def inner(request, *args, **kwargs):

            if not request.user.has_perm(perm):
                err_msg = f"Forbidden. You don't not have {perm} permission"
                if request.headers.get('x-requested-with') == 'XMLHttpRequest':
                    response = HttpResponseForbidden(
                        json.dumps({
                            'msg': err_msg
                        }),
                        content_type="application/json"
                    )
                else:
                    response = HttpResponseForbidden(err_msg)

                log_response(
                    "Access forbidden for %s to %s",
                    request.user,
                    request.path,
                    response=response,
                    request=request,
                )
                return response
            return func(request, *args, **kwargs)
        return inner
    return decorator



import json
import logging

from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import Group
from django.http import (Http404, HttpResponse, HttpResponseBadRequest,
                         HttpResponseForbidden)
from django.utils.translation import gettext as _
from papermerge.core.auth import (delete_access_perms,
                                  get_access_perms_as_hash, set_access_perms)
from papermerge.core.models import Access, BaseTreeNode, User

logger = logging.getLogger(__name__)


@login_required
def access_view(request, id):
    """
    Returns json of access right for a given node
    """
    try:
        node = BaseTreeNode.objects.get(id=id)
    except BaseTreeNode.DoesNotExist:
        raise Http404("Node does not exists")

    if not request.user.has_perm(Access.PERM_READ, node):
        return HttpResponseForbidden()

    result = []
    if request.method == 'GET':
        for acc in node.access_set.all():
            item = {}
            if acc.user:
                item['name'] = acc.user.username
                item['model'] = 'user'
            if acc.group:
                item['name'] = acc.group.name
                item['model'] = 'group'

            item['id'] = acc.id
            item['access_type'] = acc.access_type
            item['access_inherited'] = acc.access_inherited
            item['permissions'] = get_access_perms_as_hash(
                node,
                item['model'],
                item['name']
            )

            result.append(item)
    if request.method == 'POST':
        # request.is_ajax is repricated since Django 3.1
        if request.headers.get('x-requested-with') == 'XMLHttpRequest':
            if not request.user.has_perm(Access.PERM_CHANGE_PERM, node):
                return HttpResponseForbidden()
            #
            # json data of following format is expected:
            #   {
            #     'add': [
            #        {
            #           'model': 'user',
            #           'name': 'margaret',
            #           'access_type': 'allow',
            #           'permissions': {
            #               'read': true, 'write': true, 'delete': false
            #           }
            #        }
            #      ],
            #    'delete': [
            #        {
            #            'model': 'user',
            #            'name': 'uploader',
            #            'access_type': 'allow',
            #            'permissions': {  // does not matter,
            #                'read': true
            #            }
            #        }
            #   ]
            #   }
            #

            access_data = json.loads(request.body)
            if 'add' in access_data.keys():
                access_diffs = set_access_perms(
                    node,
                    access_data['add']
                )
                node.propagate_changes(
                    diffs_set=access_diffs,
                    apply_to_self=False
                )
            if 'delete' in access_data.keys():
                if len(access_data['delete']) > 0:
                    access_diffs = delete_access_perms(
                        node,
                        access_data['delete']
                    )
                    node.propagate_changes(
                        diffs_set=access_diffs,
                        apply_to_self=False
                    )
        else:  # POST but not ajax
            return HttpResponseBadRequest()

    return HttpResponse(
        json.dumps(
            {
                'access': result,
                'node_id': node.id
            }
        ),
        content_type="application/json"
    )


@login_required
def user_or_groups(request):
    """
    Returns list of user and groups to be displayed in
    permission editor for any document/folder
    """
    if not request.user.has_perm('core.view_access'):
        return HttpResponseForbidden(
            json.dumps({
                'msg': _("You don't have permissions for access model")
            }),
            content_type="application/json"
        )
    result = []
    for user in User.objects.order_by('username').all():
        item = {}
        item['model'] = 'user'
        item['name'] = user.username
        result.append(item)

    for group in Group.objects.order_by('name').all():
        item = {}
        item['model'] = 'group'
        item['name'] = group.name
        result.append(item)

    return HttpResponse(
        json.dumps({'usergroups': result}),
        content_type="application/json"
    )



from django.db import models
from django.utils.translation import ugettext_lazy as _

from polymorphic_tree.managers import (
    PolymorphicMPTTModelManager,
    PolymorphicMPTTQuerySet
)

from papermerge.core.models.diff import Diff
from papermerge.core.models.kvstore import (
    KVCompNode,
    KVNode,
    KVStoreNode,
    get_currency_formats,
    get_date_formats,
    get_kv_types,
    get_numeric_formats
)
from papermerge.core.models.node import (
    BaseTreeNode,
    RELATED_NAME_FMT,
    RELATED_QUERY_NAME_FMT
)
from papermerge.search import index


class FolderManager(PolymorphicMPTTModelManager):
    pass


class FolderQuerySet(PolymorphicMPTTQuerySet):

    def delete(self, *args, **kwargs):
        for node in self:
            descendants = node.get_descendants()

            if descendants.count() > 0:
                descendants.delete(*args, **kwargs)
            # At this point all descendants were deleted.
            # Self delete :)
            try:
                node.delete(*args, **kwargs)
            except BaseTreeNode.DoesNotExist:
                # this node was deleted by earlier recursive call
                # it is ok, just skip
                pass


CustomFolderManager = FolderManager.from_queryset(FolderQuerySet)


class Folder(BaseTreeNode, index.Indexed):

    # special folders' name always starts with a DOT (. character)
    INBOX_NAME = ".inbox"

    search_fields = [
        index.SearchField('title'),
        index.SearchField('text', partial_match=True, boost=2),
        index.SearchField('notes')
    ]

    objects = CustomFolderManager()

    def delete(self, *args, **kwargs):
        descendants = self.basetreenode_ptr.get_descendants()

        if descendants.count() > 0:
            for node in descendants:
                try:
                    node.delete(*args, **kwargs)
                except BaseTreeNode.DoesNotExist:
                    pass
        # At this point all descendants were deleted.
        # Self delete :)
        try:
            super().delete(*args, **kwargs)
        except BaseTreeNode.DoesNotExist:
            # this node was deleted by earlier recursive call
            # it is ok, just skip
            pass

    def to_dict(self):
        item = {}

        item['id'] = self.id
        item['title'] = self.title
        item['created_at'] = self.human_created_at
        item['updated_at'] = self.human_updated_at
        item['owner'] = self.user.username
        item['timestamp'] = self.created_at.timestamp()

        if self.parent:
            item['parent_id'] = self.parent.id
        else:
            item['parent_id'] = ''
        item['ctype'] = 'folder'

        tags = []
        for tag in self.tags.all():
            tags.append(tag.to_dict())
        item['tags'] = tags

        kvstore = []
        for kv in self.kv.all():
            kvstore.append(kv.to_dict())

        item['metadata'] = {}
        item['metadata']['kvstore'] = kvstore
        item['metadata']['currency_formats'] = get_currency_formats()
        item['metadata']['date_formats'] = get_date_formats()
        item['metadata']['numeric_formats'] = get_numeric_formats()
        item['metadata']['kv_types'] = get_kv_types()

        return item

    @property
    def kv(self):
        return KVNode(instance=self)

    @property
    def kvcomp(self):
        return KVCompNode(instance=self)

    def inherit_kv_from(self, folder):
        instances_set = []

        for key in folder.kv.keys():
            instances_set.append(
                KVStoreNode(key=key, kv_inherited=True, node=self)
            )

        # if there is metadata
        if len(instances_set) > 0:
            diff = Diff(
                operation=Diff.ADD,
                instances_set=instances_set
            )

            self.propagate_changes(
                diffs_set=[diff],
                apply_to_self=True
            )

    class Meta:
        verbose_name = _("Folder")
        verbose_name_plural = _("Folders")

    def __str__(self):
        return self.title


class AbstractFolder(models.Model):
    base_ptr = models.ForeignKey(
        Folder,
        related_name=RELATED_NAME_FMT,
        related_query_name=RELATED_QUERY_NAME_FMT,
        on_delete=models.CASCADE
    )

    class Meta:
        abstract = True

    def get_title(self):
        return self.base_ptr.title


def get_inbox_children(user):
    """
    Returns a ``QuerySet`` containing the immediate children nodes
    of given user's inbox folder
    """
    inbox_node = BaseTreeNode.objects.get(
        title=Folder.INBOX_NAME,
        user=user
    )

    return inbox_node.get_children()



# Generated by Django 3.0.2 on 2020-04-19 07:46

from django.db import migrations


def insert_language_maps(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0006_languagemap'),
    ]

    operations = [
        migrations.RunPython(insert_language_maps),
    ]



# Generated by Django 3.0.2 on 2020-04-19 07:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0005_auto_20200419_0639'),
    ]

    operations = [
    ]



from django.template import Library
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.urls import reverse

from papermerge.core.lib.lang import LANG_DICT
from papermerge.contrib.admin.models import LogEntry


register = Library()


def url_for_folder(node):
    return f"/#{node.id}"


def url_for_document(node):
    return f"/#{node.id}"


def build_url_for_index(
    html_class_attr='',
    title=''
):
    url = reverse('admin:index')

    link = format_html(
        '<a href="{}" class="{}" alt="{}">'
        '{}</a>',
        url,
        html_class_attr,
        title,
        title
    )

    return link


def build_url_for_node(node, html_class_attr=''):

    if node.is_folder():
        url = url_for_folder(node)
    else:
        url = url_for_document(node)

    link = format_html(
        '<a href="{}" class="{}" data-id="{}" alt="{}">'
        '{}</a>',
        url,
        html_class_attr,
        node.id,
        node.title,
        node.title
    )

    return link


def build_tree_path(
    node,
    include_self=False,
    include_index=False,
    html_class_attr=''
):
    """
    Returns an html formated path of the Node.
    Example:
        Documents > Folder A > Folder B > Document C
    Where each node is an html anchor with href to the element.
    Node is instance of core.models.BaseTreeNode.
    include_index will add url to the index of boss page.
    """
    if node:
        ancestors = node.get_ancestors(include_self=include_self)
    else:
        ancestors = []

    titles = [
        build_url_for_node(item, html_class_attr=html_class_attr)
        for item in ancestors
    ]

    if include_index:
        titles.insert(
            0,
            build_url_for_index(html_class_attr=html_class_attr)
        )

    return mark_safe(' › '.join(titles))


@register.inclusion_tag('admin/widgets/ocr_language_select.html')
def ocr_language_select(user):
    languages = []
    for key, value in LANG_DICT.items():

        lang = {}
        lang['tsr_code'] = key
        lang['human'] = value.capitalize()
        if user.preferences['ocr__OCR_Language'] == key:
            lang['selected'] = 'selected'
        else:
            lang['selected'] = ''

        languages.append(lang)

    return {'languages': languages}


@register.simple_tag(takes_context=True)
def activate_on(context, names):
    """
    names is a string of words separated by comma.
    Example:
        'group, groups'
        'users, user'

    Maybe a single word as well (without comma):
        'user'
    """
    cleaned_names = [
        name.strip() for name in names.split(',')
    ]
    if context['request'].resolver_match.url_name in cleaned_names:
        return 'active'

    return ''


@register.simple_tag
def boolean_icon(boolean_value, show_empty_on_false=False):

    icon_html = mark_safe("<i class='fa {} {}'></i>")

    if boolean_value:
        return format_html(
            icon_html,
            "fa-check",
            "text-success"
        )

    if show_empty_on_false:
        return ''

    return format_html(
        icon_html,
        "fa-times",
        "text-danger"
    )


@register.simple_tag()
def search_folder_path(node):
    return build_tree_path(
        node,
        include_self=True,
        include_index=True,
        html_class_attr="mx-1"
    )


@register.simple_tag()
def search_document_path(node):
    return build_tree_path(
        node,
        include_self=False,
        include_index=False,
        html_class_attr="mx-1"
    )


@register.simple_tag()
def tree_path(node):
    return build_tree_path(
        node,
        include_self=True,
        include_index=False,
        html_class_attr="mx-1"
    )


@register.simple_tag()
def tags_line(item):
    """
    item must have tags attribute (item.tags.all() iteratable)
    """
    li = "<li class='tag' style='color:{}; background-color:{}'>{}</li>"
    li_items = [
        format_html(
            mark_safe(
                li
            ),
            tag.fg_color,
            tag.bg_color,
            tag.name
        )
        for tag in item.tags.all()
    ]
    result = format_html(
        mark_safe(''.join(li_items))
    )

    return result


@register.filter
def log_level(level_as_int):
    """
    logging.INFO -> _("Info")
    logging.DEBUG -> _("Debug")
    etc
    """

    for level in LogEntry.LEVELS:
        if level_as_int == level[0]:
            return level[1]

    return None


@register.simple_tag(takes_context=True)
def localized_datetime(context, datetime_instance):
    user = context['request'].user

    if user:
        date_fmt = user.preferences['localization__date_format']
        time_fmt = user.preferences['localization__time_format']
        # include seconds as well
        fmt = f"{date_fmt} {time_fmt}:%S"
        ret_str = datetime_instance.strftime(fmt)

        return ret_str


@register.simple_tag(takes_context=True)
def localized_date(context, datetime_instance):
    user = context['request'].user

    if user:
        date_fmt = user.preferences['localization__date_format']
        ret_str = datetime_instance.strftime(date_fmt)

        return ret_str


@register.simple_tag(takes_context=True)
def select_if_current_version(context, version, forloop_first):
    """
    Decide if current select option should be selected or no.
    This template is called within a for loop of document versions.

    This simple tag is used in info widget of the document (if document
    has more then one version).
    If document has more then one version, a select tag will be used so that
    user can select which document version will be displayed in documet viewer.

    The fact that version X is currently displayed to the user is determined by
    URL GET parameter http://.../?version=X.
    Thus, if URL parameter version == 'X' and simple tag argument
    ``version`` == 'X' then this current option must be selected.
    In case when URL parameter version is not present at all - the last
    version will be selected. We are iterating along with first element
    if ``forloop_first`` argument is True.
    """
    request = context['request']
    param_version = request.GET.get('version', None)

    if param_version is None:
        if forloop_first:
            return 'selected'

    # here param_version is not None
    str_version = str(version)

    if str_version == param_version:
        return 'selected'

    return ''


def is_latest_version(version, versions):
    if version is None:
        return True

    version = int(version)
    if versions:
        if version == int(versions[-1]):
            return True

    return False


@register.simple_tag(takes_context=True)
def thumb_arrow_up(context):
    """
    User can change page order (move page up) only is currently
    latest version of the document is displayed
    """
    has_perm_write = context['has_perm_write']
    request = context['request']
    versions = context['versions']
    version = request.GET.get('version', None)

    ret = mark_safe(
        "<div class='arrow-up-control'>"
        "<a href='#'><i class='fa fa-arrow-circle-up'></i></a>"
        "</div>"
    )

    if has_perm_write and is_latest_version(
        version, versions
    ):
        return ret

    return ''


@register.simple_tag(takes_context=True)
def thumb_arrow_down(context):
    """
    User can change page order (move page down) only is currently
    latest version of the document is displayed
    """
    has_perm_write = context['has_perm_write']
    request = context['request']
    versions = context['versions']
    version = request.GET.get('version', None)

    ret = mark_safe(
        "<div class='arrow-down-control'>"
        "<a href='#'><i class='fa fa-arrow-circle-down'></i></a>"
        "</div>"
    )

    if has_perm_write and is_latest_version(
        version, versions
    ):
        return ret

    return ''
